import os
import gdown
from pathlib import Path
from typing import Sequence
from .data_model import Statement, Entity

BASE_PATH = ".sskb_data"
BASE_URL = "http://personalpages.manchester.ac.uk/staff/danilo.carvalho/sskb/"


class KnowledgeBase(Sequence[Statement]):
    def __init__(self, path: str, url: str, **kwargs):
        self.data_path: str = os.path.normpath(os.path.join(str(Path.home()), BASE_PATH, path))
        if (not os.path.exists(self.data_path) and url):
            os.makedirs(os.path.join(*os.path.split(self.data_path)[:-1]), exist_ok=True)
            gdown.download(url, self.data_path)
        self.id = ""
        self.entities: list[Entity] = list()

    def __len__(self):
        raise NotImplementedError

    def __getitem__(self, index):
        raise NotImplementedError

    def __iter__(self):
        raise NotImplementedError

    def keys(self):
        raise NotImplementedError

    @staticmethod
    def from_resource(locator: str):
        raise NotImplementedError

    @staticmethod
    def download_resource(path: str, url: str):
        """
        Downloads a resource from a specified URL to a specified path

        This method is used to download original or preprocessed data that backs the KnowledgeBase instances.

        Args:
            path (str): The subpath to the resource within the base path.
            url (str): The URL from which to download the resource.

        Returns:
            str: The normalized path where the resource has been downloaded.
        """
        data_path: str = os.path.normpath(os.path.join(str(Path.home()), BASE_PATH, path))
        if (not os.path.exists(data_path)):
            os.makedirs(os.path.join(*os.path.split(data_path)[:-1]), exist_ok=True)
            gdown.download(url, data_path)

        return data_path






from zipfile import ZipFile
from io import TextIOWrapper
from csv import DictReader
from tqdm import tqdm
from spacy.lang.en import English
from saf import Token, Sentence
from ..skb import KnowledgeBase
from ..data_model import Statement, Entity

PATH = "WorldTree/tg2020task-dataset.zip"
URL = "https://cognitiveai.org/dist/tg2020task-dataset.zip"


class WorldTreeKB(KnowledgeBase):
    """
    Wrapper for the WorldTree dataset (Jansen et al., 2020): https://github.com/cognitiveailab/tg2020task
    """
    def __init__(self, path: str = PATH, url: str = URL):
        super().__init__(path, url)
        self.tokenizer = English().tokenizer
        self.id = PATH.split(".")[0]
        self.key_idx: dict[str, list[Statement]] = dict()
        if (not url):
            return

        ent_nameset = set()
        num_facts = 0
        with ZipFile(self.data_path) as dataset_file:
            self.data = list()
            tbl_paths = [fpath for fpath in dataset_file.namelist() if fpath.startswith("tables/")]
            for tbl_path in tqdm(tbl_paths, desc="Loading fact tables"):
                table_name = tbl_path.split("/")[-1].split(".")[0]
                with TextIOWrapper(dataset_file.open(tbl_path, "r"), encoding="utf-8") as tbl_file:
                    reader = DictReader(tbl_file, delimiter="\t")
                    for row in reader:
                        uid = row["[SKIP] UID"].strip()
                        self.key_idx[uid] = list()
                        terms = [(row[field], field) for field in row if "SKIP" not in field]
                        surface = " ".join([t[0] for t in terms if t[0].strip()])
                        stt = Statement(surface)
                        stt.annotations["UID"] = uid
                        stt.annotations["SSKB_ID"] = f"{self.id}::{uid}"
                        stt.annotations["type"] = "fact"
                        stt.annotations["table"] = table_name
                        if ("[SKIP] DEP" in row and row["[SKIP] DEP"].startswith("SW")):
                            stt.annotations["dep"] = row["[SKIP] DEP"].strip().split()[1:]

                        for t in terms:
                            term = Sentence()
                            term.surface = t[0]
                            for tok in self.tokenizer(t[0]):
                                token = Token()
                                token.surface = tok.text
                                term.tokens.append(token)

                            term.annotations["role"] = t[1]
                            stt.terms.append(term)
                            stt.tokens.extend(term.tokens)

                            if ("OBJECT" in t[1]):
                                ent_names = [e.strip() for e in t[0].split(";")]
                                for name in ent_names:
                                    if (name not in ent_nameset):
                                        ent = Entity(name, self.id)
                                        ent.metadata["table"] = {table_name}
                                        self.entities.append(ent)
                                        stt.entities.append(ent)
                                        ent_nameset.add(name)
                                    else:
                                        ent = [e for e in self.entities if e.surface == name][0]
                                        ent.metadata["table"].add(table_name)

                        if (stt.surface.strip()):
                            self.data.append(stt)
                            self.key_idx[uid].append(stt)
                            num_facts += 1

            for split in ["train", "dev", "test"]:
                with TextIOWrapper(dataset_file.open(f"questions.{split}.tsv"), encoding="utf-8") as q_file:
                    reader = DictReader(q_file, delimiter="\t")
                    for row in tqdm(reader, desc=f"Loading questions [{split}]"):
                        stt = Statement(row["question"])
                        stt.annotations["SSKB_ID"] = f"{self.id}::{row['QuestionID']}"
                        for field in row:
                            stt.annotations[field] = row[field]
                        del stt.annotations["question"]
                        stt.annotations["type"] = "question"
                        stt.annotations["split"] = split
                        stt.annotations["explanation"] = [
                            tuple(exp.split("|"))
                            for exp in stt.annotations["explanation"].split()
                        ]
                        for expl in stt.annotations["explanation"]:
                            prem = [p for p in self.data[:num_facts] if (p.annotations["UID"] == expl[0])]
                            if (prem):
                                stt.premises.append(prem[0])
                                self.key_idx[expl[0]].append(stt)

                        for tok in self.tokenizer(stt.surface):
                            token = Token()
                            token.surface = tok.text
                            stt.tokens.append(token)

                        if (stt.surface.strip()):
                            self.data.append(stt)

    def __iter__(self):
        return iter(self.data)

    def __len__(self):
        return len(self.data)

    def __getitem__(self, idx: int | str | slice) -> Statement | list[Statement]:
        """Fetches the ith statement in the KB.

        Args:
            idx (int): index for the ith term in the KB.

        :return: A single term definition (Statement).
        """

        if (isinstance(idx, int) or isinstance(idx, slice)):
            item = self.data[idx]
        else:
            item = self.key_idx[idx]

        return item

    def keys(self):
        return self.key_idx.keys()




import json
import pickle
import gzip
from zipfile import ZipFile
from io import TextIOWrapper
from tqdm import tqdm
from spacy.lang.en import English
from saf import Token
from ..skb import KnowledgeBase, BASE_URL
from ..data_model import Statement, Entity

PATH = "ProofWiki/proofwiki.zip"
URL = BASE_URL + "proofwiki.zip"

ANNOT_RESOURCES = {
    "proofwiki[entities]": {
        "path": "ProofWiki/proofwiki[entities].pickle.gz",
        "url": BASE_URL + "proofwiki[entities].pickle.gz"
    }
}


class ProofWikiKB(KnowledgeBase):
    """
    Wrapper for the ProofWiki dataset (Ferreira et al., 2018): https://github.com/ai-systems/tg2022task_premise_retrieval
    """
    def __init__(self, path: str = PATH, url: str = URL):
        super().__init__(path, url)
        self.tokenizer = English().tokenizer
        self.id = PATH.split(".")[0]
        self.key_idx: dict[int, list[Statement]] = dict()
        if (not url):
            return

        with ZipFile(self.data_path) as dataset_file:
            self.data = list()

            with TextIOWrapper(dataset_file.open("ProofWiki/def_titles.txt"), encoding="utf-8") as ent_file:
                for line in ent_file:
                    ent = Entity(line.strip(), self.id)
                    self.entities.append(ent)

            with TextIOWrapper(dataset_file.open("ProofWiki/knowledge_base.json"), encoding="utf-8") as data_file:
                data = json.load(data_file)
                for key in tqdm(data, desc=f"Loading data [Knowledge Base]"):
                    self.key_idx[int(key)] = list()
                    stt = Statement(data[key].strip())
                    stt.annotations["split"] = "KB"
                    stt.annotations["type"] = "fact"
                    stt.annotations["id"] = int(key)
                    stt.annotations["SSKB_ID"] = f"{self.id}::{key}"

                    for tok in self.tokenizer(stt.surface):
                        token = Token()
                        token.surface = tok.text
                        stt.tokens.append(token)

                    self.data.append(stt)
                    self.key_idx[int(key)].append(stt)

            for split in ["train", "dev", "test"]:
                with TextIOWrapper(dataset_file.open(f"ProofWiki/{split}_set.json"), encoding="utf-8") as data_file:
                    data = json.load(data_file)
                    for key in tqdm(data, desc=f"Loading data [{split}]"):
                        self.key_idx[int(key)] = list()
                        stt = Statement(data[key]["text"].strip())
                        stt.annotations["split"] = split
                        stt.annotations["type"] = "proposition"
                        stt.annotations["id"] = int(key)
                        stt.annotations["SSKB_ID"] = f"{self.id}::{key}"
                        for prem_id in data[key]["premises"]:
                            if (prem_id in self.key_idx):
                                stt.premises.extend(self.key_idx[prem_id])

                        for tok in self.tokenizer(stt.surface):
                            token = Token()
                            token.surface = tok.text
                            stt.tokens.append(token)

                        self.data.append(stt)
                        self.key_idx[int(key)].append(stt)

        stt_text_l = [stt.surface.lower() for stt in self.data]
        for ent in tqdm(self.entities, desc="Searching entities"):
            ent_name = ent.surface.lower()
            stt_matches = [stt_idx for stt_idx, stt_txt in enumerate(stt_text_l)
                           if (ProofWikiKB.entity_name_search(ent_name,stt_txt))]
            for stt_idx in stt_matches:
                self.data[stt_idx].entities.append(ent)

    def __iter__(self):
        return iter(self.data)

    def __len__(self):
        return len(self.data)

    def __getitem__(self, idx: int | slice) -> Statement | list[Statement]:
        """Fetches the ith statement in the KB.

        Args:
            idx (int): index for the ith term in the KB.

        :return: A single term definition (Statement).
        """
        if (isinstance(idx, slice)):
            item = self.data[idx]
        else:
            item = self.data[idx] if (idx < len(self)) else self.key_idx[idx]

        return item

    def keys(self):
        return self.key_idx.keys()

    @staticmethod
    def entity_name_search(ent_name: str, text: str) -> bool:
        if (ent_name in text):
            match = f" {ent_name} " in text or f"'{ent_name}'" in text or f'"{ent_name}"' in text
        else:
            match = False

        return match

    @staticmethod
    def from_resource(locator: str):
        """
        Downloads a pre-annotated resource available at the specified locator

        Example:
            >>> kb = ProofWikiKB.from_resource("proofwiki[entities]")
        """
        kb = None
        if (locator in ANNOT_RESOURCES):
            path = ANNOT_RESOURCES[locator]["path"]
            url = ANNOT_RESOURCES[locator]["url"]
            data_path = KnowledgeBase.download_resource(path, url)
            with gzip.open(data_path, "rb") as resource_file:
                kb = pickle.load(resource_file)
        else:
            print(f"No resource found at locator: {locator}")

        return kb





import os
import sys
import pkgutil
import pyclbr
import importlib

for mod in pkgutil.iter_modules([os.path.abspath(os.path.dirname(__file__))]):
    clbr = pyclbr.readmodule(mod.name, [mod.module_finder.path, mod.name])
    for cls in clbr:
        if (clbr[cls].super and clbr[cls].super[0] == "KnowledgeBase"):
            imp_cls = importlib.import_module(f".{mod.name}", __name__).__dict__[clbr[cls].name]
            sys.modules[__name__].__setattr__(clbr[cls].name, imp_cls)

del clbr
del cls
del imp_cls
del mod



from zipfile import ZipFile
from io import TextIOWrapper
from csv import DictReader
from tqdm import tqdm
from spacy.lang.en import English
from saf import Token
from ..skb import KnowledgeBase, BASE_URL
from ..data_model import Statement

PATH = "eSNLI/eSNLI.zip"
URL = BASE_URL + "eSNLI.zip"


class ESNLIKB(KnowledgeBase):
    """
    Wrapper for the e-SNLI dataset (Camburu et al., 2018): https://github.com/OanaMariaCamburu/e-SNLI
    """
    def __init__(self, path: str = PATH, url: str = URL):
        super().__init__(path, url)
        self.tokenizer = English().tokenizer
        self.id = PATH.split(".")[0]
        self.key_idx: dict[str, list[Statement]] = dict()
        if (not url):
            return

        with ZipFile(self.data_path) as dataset_file:
            self.data = list()

            for split in ["train", "dev", "test"]:
                with TextIOWrapper(dataset_file.open(f"eSNLI/esnli_{split}.tsv"), encoding="utf-8") as data_file:
                    reader = DictReader(data_file, delimiter="\t")
                    for row in tqdm(reader, desc=f"Loading data [{split}]"):
                        sents = list()
                        self.key_idx[row["pairID"]] = list()
                        for sent_key in ["Sentence1", "Sentence2"]:
                            sent = row[sent_key]
                            stt = Statement(sent)
                            stt.annotations["split"] = split
                            stt.annotations["label"] = row["gold_label"]
                            stt.annotations["type"] = "fact"
                            stt.annotations["number"] = int(sent_key[-1])
                            stt.annotations["id"] = row["pairID"]
                            stt.annotations["SSKB_ID"] = f"{self.id}::{row['pairID']}"
                            for i in range(1, 4):
                                if (f"{sent_key}_marked_{i}" in row):
                                    stt.annotations[f"{sent_key}_marked_{i}"] = row[f"{sent_key}_marked_{i}"]
                                    stt.annotations[f"{sent_key}_Highlighted_{i}"] = row[f"{sent_key}_Highlighted_{i}"]

                            for tok in self.tokenizer(stt.surface):
                                token = Token()
                                token.surface = tok.text
                                stt.tokens.append(token)

                            self.data.append(stt)
                            self.key_idx[row["pairID"]].append(stt)
                            sents.append(stt)

                        for expl_key in ["Explanation_1", "Explanation_2", "Explanation_3"]:
                            if (expl_key not in row):
                                continue

                            expl = row[expl_key]
                            for expl_sent in expl.split(". "):
                                stt = Statement(expl_sent)
                                stt.annotations["split"] = split
                                stt.annotations["label"] = row["gold_label"]
                                stt.annotations["type"] = "explanation"
                                stt.annotations["number"] = expl_key[-1]
                                stt.annotations["id"] = row["pairID"]
                                stt.annotations["SSKB_ID"] = f"{self.id}::{row['pairID']}"

                                for tok in self.tokenizer(stt.surface):
                                    token = Token()
                                    token.surface = tok.text
                                    stt.tokens.append(token)

                                for sent in sents:
                                    sent.premises.append(stt)

                                self.data.append(stt)
                                self.key_idx[row["pairID"]].append(stt)

    def __iter__(self):
        return iter(self.data)

    def __len__(self):
        return len(self.data)

    def __getitem__(self, idx: int | str | slice) -> Statement | list[Statement]:
        """Fetches the ith statement in the KB.

        Args:
            idx (int): index for the ith term in the KB.

        :return: A single term definition (Statement).
        """
        if (isinstance(idx, int) or isinstance(idx, slice)):
            item = self.data[idx]
        else:
            item = self.key_idx[idx]

        return item

    def keys(self):
        return self.key_idx.keys()




from typing import Self, Any
from dataclasses import dataclass, field
from xxhash import xxh128_hexdigest
from saf import Sentence


@dataclass
class Entity:
    surface: str
    kbid: str
    metadata: dict[str, Any] = field(default_factory=dict)
    synonyms: list[Self] = field(default_factory=list)
    hypernyms: list[Self] = field(default_factory=list)


class Statement(Sentence):
    def __init__(self, surface: str):
        super().__init__()
        self.surface: str = surface
        self.premises: list[Self] = list()
        self.entities: list[Entity] = list()

    @property
    def id(self) -> str:
        return self.annotations["SSKB_ID"] if ("SSKB_ID" in self.annotations) else xxh128_hexdigest(self.surface)





import json
from zipfile import ZipFile
from io import TextIOWrapper
from tqdm import tqdm
from spacy.lang.en import English
from saf import Token
from ..skb import KnowledgeBase
from ..data_model import Statement

PATH = "NLI4CT/Complete_dataset.zip"
URL = "https://github.com/ai-systems/nli4ct/raw/main/Complete_dataset.zip"


class NLI4CTKB(KnowledgeBase):
    """
    Wrapper for the NLI4CT dataset (Julien et al., 2018): https://github.com/ai-systems/nli4ct
    """
    def __init__(self, path: str = PATH, url: str = URL):
        super().__init__(path, url)
        self.tokenizer = English().tokenizer
        self.id = PATH.split(".")[0]
        self.key_idx: dict[str, list[Statement]] = dict()
        if (not url):
            return

        with ZipFile(self.data_path) as dataset_file:
            self.data = list()

            ct_fpaths = [path for path in dataset_file.namelist()
                         if (path.startswith("Complete_dataset/CT json") and path.endswith(".json"))]
            for fpath in tqdm(ct_fpaths, desc="Loading CT data"):
                with TextIOWrapper(dataset_file.open(fpath), encoding="utf-8") as ct_file:
                    ct_data = json.load(ct_file)
                    ct_id = ct_data["Clinical Trial ID"]
                    self.key_idx[ct_id] = list()
                    del ct_data["Clinical Trial ID"]
                    for section in ct_data:
                        idx = 0
                        for line in ct_data[section]:
                            stt = Statement(line.strip())
                            stt.annotations["id"] = ct_id
                            stt.annotations["section"] = section
                            stt.annotations["index"] = idx
                            stt.annotations["SSKB_ID"] = f"{self.id}::{ct_id}"

                            for tok in self.tokenizer(stt.surface):
                                token = Token()
                                token.surface = tok.text
                                stt.tokens.append(token)

                            self.data.append(stt)
                            self.key_idx[ct_id].append(stt)
                            idx += 1

            for split in ["train", "dev", "Gold_test"]:
                with TextIOWrapper(dataset_file.open(f"Complete_dataset/{split}.json"), encoding="utf-8") as data_file:
                    data = json.load(data_file)
                    for uuid in tqdm(data, desc=f"Loading data [{split}]"):
                        self.key_idx[uuid] = list()
                        stt = Statement(data[uuid]["Statement"])
                        stt.annotations["id"] = uuid
                        stt.annotations["split"] = split
                        stt.annotations["type"] = data[uuid]["Type"]
                        stt.annotations["section_id"] = data[uuid]["Section_id"]
                        stt.annotations["primary_id"] = data[uuid]["Primary_id"]
                        stt.annotations["SSKB_ID"] = f"{self.id}::{uuid}"
                        if ("Label" in data[uuid]):
                            stt.annotations["label"] = data[uuid]["Label"]

                        for tok in self.tokenizer(stt.surface):
                            token = Token()
                            token.surface = tok.text
                            stt.tokens.append(token)

                        for prem_idx in data[uuid]["Primary_evidence_index"]:
                            prem = [
                                p for p in self.key_idx[data[uuid]["Primary_id"]]
                                if (p.annotations["section"] == data[uuid]["Section_id"] and
                                    p.annotations["index"] == prem_idx)
                            ]
                            stt.premises.append(prem[0])

                        if ("Secondary_id" in data[uuid]):
                            stt.annotations["secondary_id"] = data[uuid]["Secondary_id"]
                            for prem_idx in data[uuid]["Secondary_evidence_index"]:
                                prem = [
                                    p for p in self.key_idx[data[uuid]["Secondary_id"]]
                                    if (p.annotations["section"] == data[uuid]["Section_id"] and
                                        p.annotations["index"] == prem_idx)
                                ][0]
                                stt.premises.append(prem)

                        self.data.append(stt)
                        self.key_idx[uuid].append(stt)

    def __iter__(self):
        return iter(self.data)

    def __len__(self):
        return len(self.data)

    def __getitem__(self, idx: int | str | slice) -> Statement | list[Statement]:
        """Fetches the ith statement in the KB.

        Args:
            idx (int): index for the ith term in the KB.

        :return: A single term definition (Statement).
        """
        if (isinstance(idx, slice) or isinstance(idx, int)):
            item = self.data[idx]
        else:
            item = self.key_idx[idx]

        return item

    def keys(self):
        return self.key_idx.keys()



